(* $Id: rpc_program.ml,v 1.1 1999/04/12 01:23:45 gerd Exp $
 * ----------------------------------------------------------------------
 *
 *)

(* The concept of RPC "programs" *)

open Rtypes
open Xdr

type t =
    { prog_nr : uint4;
      vers_nr : uint4;
      spec_ts : xdr_type_system;
      spec_procs : 
	( string * (uint4 * xdr_type_term * xdr_type_term)) list;
      mutable spec_procs_validated : 
	( string * (uint4 * xdr_type * xdr_type)) list
    }

let create prognr versnr ts procs =
  { prog_nr = prognr;
    vers_nr = versnr;
    spec_ts = ts;
    spec_procs = procs;
    spec_procs_validated = []
  } 

let program_number p = p.prog_nr
    
let version_number p = p.vers_nr

let signature p procname =
  try
    List.assoc procname p.spec_procs_validated
  with
    Not_found ->
      let (m,s,t) = List.assoc procname p.spec_procs in
      let s_type = expanded_xdr_type p.spec_ts s in
      let t_type = expanded_xdr_type p.spec_ts t in
      p.spec_procs_validated <- 
	(procname,(m,s_type,t_type)) :: p.spec_procs_validated;
      (m,s_type, t_type)
	
let procedure_number p procname =
  let (m,s,t) = List.assoc procname p.spec_procs in
  m

(* ======================================================================
 * History:
 * 
 * $Log: rpc_program.ml,v $
 * Revision 1.1  1999/04/12 01:23:45  gerd
 * 	Initial revision.
 *
 * 
 *)
