(*
 * $Id: rtypes.ml,v 1.1 1999/04/04 20:04:56 gerd Exp $
 *
 *)

(* NOTE: Parts of this implementation depend very much of representation 
 * details of O'Caml 2.02. It is not guaranteed that this works in future
 * versions of O'Caml as well.
 *)

(* representation types *)

type int1 = string;;    (* all types are opaque *)
type int2 = string;;
type int4 = string;;
type int8 = string;;

type uint1 = string;;
type uint2 = string;;
type uint4 = string;;
type uint8 = string;;

type fp4 = string;;    (* IEEE representation of fp numbers *)
type fp8 = string;;

(* Note: In OCAML, only int1, int2, uint1 and uint2 can be
 * mapped to OCAML's own int type. OCAML's integers have only
 * 31 bits (including sign)
 *)

exception Cannot_represent;;
(* raised if numbers are too big to map them to other type *)

(**********************************************************************)
(* mk_[u]intn                                                         *)
(**********************************************************************)

let mk_int1 c0 =
  String.make 1 c0
;;

let mk_int2 (c1,c0) =
  let s = String.make 2 ' ' in
  s.[0] <- c1;
  s.[1] <- c0;
  s
;;

let mk_int4 (c3,c2,c1,c0) =
  let s = String.make 4 ' ' in
  s.[0] <- c3;
  s.[1] <- c2;
  s.[2] <- c1;
  s.[3] <- c0;
  s
;;

let mk_int8 (c7,c6,c5,c4,c3,c2,c1,c0) =
  let s = String.make 4 ' ' in
  s.[0] <- c7;
  s.[1] <- c6;
  s.[2] <- c5;
  s.[3] <- c4;
  s.[4] <- c3;
  s.[5] <- c2;
  s.[6] <- c1;
  s.[7] <- c0;
  s
;;

let mk_uint1 = mk_int1;;
let mk_uint2 = mk_int2;;
let mk_uint4 = mk_int4;;
let mk_uint8 = mk_int8;;

(**********************************************************************)
(* dest_[u]intn                                                       *)
(**********************************************************************)

let dest_int1 x =
  x.[0]
;;


let dest_int2 x =
  x.[0], x.[1]
;;


let dest_int4 x =
  x.[0], x.[1], x.[2], x.[3]
;;


let dest_int8 x =
  x.[0], x.[1], x.[2], x.[3], x.[4], x.[5], x.[6], x.[7]
;;


let dest_uint1 = dest_int1;;
let dest_uint2 = dest_int2;;
let dest_uint4 = dest_int4;;
let dest_uint8 = dest_int8;;

(**********************************************************************)
(* [u]intn_as_string                                                  *)
(**********************************************************************)

let int1_as_string s = s;;
let int2_as_string s = s;;
let int4_as_string s = s;;
let int8_as_string s = s;;
let uint1_as_string s = s;;
let uint2_as_string s = s;;
let uint4_as_string s = s;;
let uint8_as_string s = s;;

(**********************************************************************)
(* int_of_[u]intn                                                     *)
(**********************************************************************)

let int_of_int1 x =
  let a = Char.code (x.[0]) in
  if a >= 128 then
    -256+a
  else
    a
;;


let int_of_uint1 x =
  let a = Char.code (x.[0]) in
    a
;;


let int_of_int2 x =
  let a0 = Char.code (x.[1]) in
  let a1 = Char.code (x.[0]) in
  if a1 >= 128 then
    -65536+((a1 lsl 8) lor a0)
  else
    (a1 lsl 8) lor a0
;;


let int_of_uint2 x =
  let a0 = Char.code (x.[1]) in
  let a1 = Char.code (x.[0]) in
    (a1 lsl 8) lor a0
;;


let int_of_int4 =
  if Sys.word_size = 32 then
    fun x ->
      let a0 = Char.code (x.[3]) in
      let a1 = Char.code (x.[2]) in
      let a2 = Char.code (x.[1]) in
      let a3 = Char.code (x.[0]) in
      if a3 >= 64 & a3 < 192 then
	raise Cannot_represent
      else
  	if a3 >= 128 then
	  lnot (((a3 lxor 255) lsl 24) lor ((a2 lxor 255) lsl 16) 
		  lor ((a1 lxor 255) lsl 8) lor (a0 lxor 255))
  	else
	  (a3 lsl 24) lor (a2 lsl 16) lor (a1 lsl 8) lor a0
  else if Sys.word_size = 64 then
    fun x ->
      let a0 = Char.code (x.[3]) in
      let a1 = Char.code (x.[2]) in
      let a2 = Char.code (x.[1]) in
      let a3 = Char.code (x.[0]) in
      if a3 >= 128 then
	lnot (((a3 lxor 255) lsl 24) lor ((a2 lxor 255) lsl 16) 
		lor ((a1 lxor 255) lsl 8) lor (a0 lxor 255))
      else
	(a3 lsl 24) lor (a2 lsl 16) lor (a1 lsl 8) lor a0
  else
    failwith "word size not supported"
;;


let int_of_uint4 =
  if Sys.word_size = 32 then
    fun x ->
      let a0 = Char.code (x.[3]) in
      let a1 = Char.code (x.[2]) in
      let a2 = Char.code (x.[1]) in
      let a3 = Char.code (x.[0]) in
      if a3 >= 64 then
	raise Cannot_represent
      else
	(a3 lsl 24) lor (a2 lsl 16) lor (a1 lsl 8) lor a0
  else if Sys.word_size = 64 then
    fun x ->
      let a0 = Char.code (x.[3]) in
      let a1 = Char.code (x.[2]) in
      let a2 = Char.code (x.[1]) in
      let a3 = Char.code (x.[0]) in
        (a3 lsl 24) lor (a2 lsl 16) lor (a1 lsl 8) lor a0
  else
    failwith "word size not supported"
;;


let int_of_int8 =
  if Sys.word_size = 32 then
    fun x ->
      let a0 = Char.code (x.[7]) in
      let a1 = Char.code (x.[6]) in
      let a2 = Char.code (x.[5]) in
      let a3 = Char.code (x.[4]) in
      let a4 = Char.code (x.[3]) in
      let a5 = Char.code (x.[2]) in
      let a6 = Char.code (x.[1]) in
      let a7 = Char.code (x.[0]) in
      if (a3 <= 63 & a4=0 & a5=0 & a6=0 & a7=0) then
	  (a3 lsl 24) lor (a2 lsl 16) lor (a1 lsl 8) lor a0
      else if (a3 >= 192 & a4=255 & a5=255 & a6=255 & a7=255) then
	  lnot (((a3 lxor 255) lsl 24) lor ((a2 lxor 255) lsl 16) 
		  lor ((a1 lxor 255) lsl 8) lor (a0 lxor 255))
      else
	raise Cannot_represent
  else if Sys.word_size = 64 then
    fun x ->
      let a0 = Char.code (x.[7]) in
      let a1 = Char.code (x.[6]) in
      let a2 = Char.code (x.[5]) in
      let a3 = Char.code (x.[4]) in
      let a4 = Char.code (x.[3]) in
      let a5 = Char.code (x.[2]) in
      let a6 = Char.code (x.[1]) in
      let a7 = Char.code (x.[0]) in
      if a7 <= 63 then
	  (a7 lsl 56) lor (a6 lsl 48) lor (a5 lsl 40) lor (a4 lsl 32) lor
	  (a3 lsl 24) lor (a2 lsl 16) lor (a1 lsl 8) lor a0
      else if a7 >= 192 then
	lnot
	  (((a7 lxor 255) lsl 56) lor 
	   ((a6 lxor 255) lsl 48) lor 
	   ((a5 lxor 255) lsl 40) lor
 	   ((a4 lxor 255) lsl 32) lor
	   ((a3 lxor 255) lsl 24) lor 
 	   ((a2 lxor 255) lsl 16) lor 
	   ((a1 lxor 255) lsl 8)  lor 
	   (a0 lxor 255))
      else 
	raise Cannot_represent
  else
    failwith "word size not supported"
;;


let int_of_uint8 =
  if Sys.word_size = 32 then
    fun x ->
      let a0 = Char.code (x.[7]) in
      let a1 = Char.code (x.[6]) in
      let a2 = Char.code (x.[5]) in
      let a3 = Char.code (x.[4]) in
      let a4 = Char.code (x.[3]) in
      let a5 = Char.code (x.[2]) in
      let a6 = Char.code (x.[1]) in
      let a7 = Char.code (x.[0]) in
      if (a3 <= 63 & a4=0 & a5=0 & a6=0 & a7=0) then
	  (a3 lsl 24) lor (a2 lsl 16) lor (a1 lsl 8) lor a0
      else
	raise Cannot_represent
  else if Sys.word_size = 64 then
    fun x ->
      let a0 = Char.code (x.[7]) in
      let a1 = Char.code (x.[6]) in
      let a2 = Char.code (x.[5]) in
      let a3 = Char.code (x.[4]) in
      let a4 = Char.code (x.[3]) in
      let a5 = Char.code (x.[2]) in
      let a6 = Char.code (x.[1]) in
      let a7 = Char.code (x.[0]) in
      if a7 <= 63 then
	  (a7 lsl 56) lor (a6 lsl 48) lor (a5 lsl 40) lor (a4 lsl 32) lor
	  (a3 lsl 24) lor (a2 lsl 16) lor (a1 lsl 8) lor a0
      else 
	raise Cannot_represent
  else
    failwith "word size not supported"
;;

(**********************************************************************)
(* intn_of_int                                                        *)
(**********************************************************************)

let int1_of_int i =
  if (i >= -128 & i < 127) then
    let a0 = i land 255 in
    mk_int1 (Char.chr a0)
  else
    failwith "int1_of_int"
;;


let uint1_of_int i =
  if (i >= 0 & i < 255) then
    let a0 = i land 255 in
    mk_uint1 (Char.chr a0)
  else
    failwith "uint1_of_int"
;;


let int2_of_int i =
  if (i >= -32768 & i <= 32767) then
    let a0 = i land 255 in
    let a1 = (i lsr 8) land 255 in
    mk_int2 (Char.chr a1, Char.chr a0)
  else
    failwith "int2_of_int"
;;


let uint2_of_int i =
  if (i >= 0 & i < 65536 ) then
    let a0 = i land 255 in
    let a1 = (i lsr 8) land 255 in
    mk_uint2 (Char.chr a1, Char.chr a0)
  else
    failwith "uint2_of_int"
;;


let int4_of_int i =
  let j = i asr 31 in
  if j = 0 or j = (-1) then
    let a0 = i land 255 in
    let a1 = (i lsr 8) land 255 in
    let a2 = (i lsr 16) land 255 in
    let a3 = (i lsr 24) land 255 in
    mk_int4 (Char.chr a3, Char.chr a2, Char.chr a1, Char.chr a0)
  else
    failwith "int4_of_int"
;;


let uint4_of_int i =
  let j = i asr 31 in      
  (* Note: There is a bug in the 32 bit implementation: i asr 32 yields
   * always i. As a workaround I use i asr 31
   *)
  if j = 0 or j = 1 then
    let a0 = i land 255 in
    let a1 = (i lsr 8) land 255 in
    let a2 = (i lsr 16) land 255 in
    let a3 = (i lsr 24) land 255 in
    mk_uint4 (Char.chr a3, Char.chr a2, Char.chr a1, Char.chr a0)
  else
    failwith "uint4_of_int"
;;


let int8_of_int i =
  let a0 = i land 255 in
  let a1 = (i lsr 8) land 255 in
  let a2 = (i lsr 16) land 255 in
  let a3 = (i lsr 24) land 255 in
  let a4 = (i lsr 32) land 255 in
  let a5 = (i lsr 40) land 255 in
  let a6 = (i lsr 48) land 255 in
  let a7 = (i lsr 56) land 255 in
  mk_int8 (Char.chr a7, Char.chr a6, Char.chr a5, Char.chr a4,
	   Char.chr a3, Char.chr a2, Char.chr a1, Char.chr a0)
;;


let uint8_of_int i =
  if i>= 0 then
    let a0 = i land 255 in
    let a1 = (i lsr 8) land 255 in
    let a2 = (i lsr 16) land 255 in
    let a3 = (i lsr 24) land 255 in
    let a4 = (i lsr 32) land 255 in
    let a5 = (i lsr 40) land 255 in
    let a6 = (i lsr 48) land 255 in
    let a7 = (i lsr 56) land 255 in
    mk_uint8 (Char.chr a7, Char.chr a6, Char.chr a5, Char.chr a4,
	      Char.chr a3, Char.chr a2, Char.chr a1, Char.chr a0)
  else
    failwith "uint8_of_int"
;;


(**********************************************************************)
(* floating-point numbers                                             *)
(**********************************************************************)

external raw_blit : string -> int -> string -> int -> int -> unit
    = "blit_string"
;;


let cast_float_to_string (x : float) =
  let s = String.create 8 in
  raw_blit (Obj.magic x : string) 0 s 0 8;
    (* raw_blit: works even with "strings" that have an incorrect 
     * representation
     *)
  s
;;


let is_little_endian =
  lazy (Char.code ((cast_float_to_string (-.1.0)).[7]) <> 0)
;;


let swap_fp8 s =
  let t = String.create 8 in
  t.[0] <- s.[7];
  t.[1] <- s.[6];
  t.[2] <- s.[5];
  t.[3] <- s.[4];
  t.[4] <- s.[3];
  t.[5] <- s.[2];
  t.[6] <- s.[1];
  t.[7] <- s.[0];
  t
;;


let float_of_fp4 x =
  failwith "not implemented"
;;


let float_of_fp8 x =
  (* CAUTION! This is magic code -- hacked... *)

  (* x is a string. Strings and double fp numbers are represented very similar.
   * So we cast x to a 'double' and add 0.0 to get all right.
   *
   * ASSUMED:
   * - The fp representation of your computer follows IEEE.
   * - We know exactly the internal representation of O'Caml.
   *)

  let x' = if Lazy.force is_little_endian then swap_fp8 x else x in
  (Obj.magic x' : float) +. 0.0
;;


let fp4_of_float x =
  failwith "not implemented"
;;

let fp8_of_float x =
  let s = cast_float_to_string x in
  if Lazy.force is_little_endian then swap_fp8 s else s
;;

let mk_fp4 (c3,c2,c1,c0) =
  failwith "not implemented"
;;

let mk_fp8 (c7,c6,c5,c4,c3,c2,c1,c0) =
  let s = String.create 8 in
  s.[0] <- c7;
  s.[1] <- c6;
  s.[2] <- c5;
  s.[3] <- c4;
  s.[4] <- c3;
  s.[5] <- c2;
  s.[6] <- c1;
  s.[7] <- c0;
  s
;;

let dest_fp4 x =
  failwith "not implemented"
;;

let dest_fp8 x =
  x.[0],
  x.[1],
  x.[2],
  x.[3],
  x.[4],
  x.[5],
  x.[6],
  x.[7]
;;


let fp4_as_string s = s;;
let fp8_as_string s = s;;

(* ======================================================================
 *
 * $Log: rtypes.ml,v $
 * Revision 1.1  1999/04/04 20:04:56  gerd
 * 	Initial release in this depot. This files have been restructured;
 * some missing features have been added (fp8 code, recursive xdr types).
 *
 *)
