open Rtypes;;
open Xdr;;
open Rpc;;

(**********************************************************************)
(*                            The port mapper                         *)
(**********************************************************************)

(* TODO: *)
(* We can currently not call the DUMP procedure because "pmaplist"
 * is not a legal type. 
 * First we must introduce recursive types.
 *)

let pm2_ts =
  [ "mapping",        X_struct [ "prog", X_uint;
				 "vers", X_uint;
				 "prot", X_uint;
				 "port", X_uint
			       ];
    "pmaplist",       X_rec("pmaplist",
                        (x_optional
			   (X_struct [ "map",  X_type "mapping";
				       "next", X_refer "pmaplist" 
				     ])));
    "call_args",      X_struct [ "prog", X_uint;
				 "vers", X_uint;
				 "proc", X_uint;
				 "args", x_opaque_max
			       ];
    "call_results",   X_struct [ "port", X_uint;
				 "res",  x_opaque_max
			       ]
  ]
;;


let pm2_spec =
  Rpc_program.create 
    (uint4_of_int 100000)
    (uint4_of_int 2)
    (validate_xdr_type_system pm2_ts)
    [ "NULL",    ((uint4_of_int 0), X_void,           X_void);
      "SET",     ((uint4_of_int 1), X_type "mapping", x_bool);
      "UNSET",   ((uint4_of_int 2), X_type "mapping", x_bool);
      "GETPORT", ((uint4_of_int 3), X_type "mapping", X_uint);
      "DUMP",    ((uint4_of_int 4), X_void, X_type "pmaplist");
      "CALLIT",  ((uint4_of_int 5), X_type "call_args", X_type "call_result")
    ] 
;;


let exec_pm2 proc arg server =
  let reply = ref XV_void in
  let esys = Unixqueue.create_unix_event_system () in
  let cl = 
    Rpc_client.create esys (Rpc_client.Inet (server, 111)) Tcp pm2_spec
  in

  Rpc_client.set_exception_handler
    cl
    (fun x -> prerr_endline ("exception: " ^ Printexc.to_string x));
(*  for i = 1 to 50 do *)
    Rpc_client.add_call cl proc arg (fun f -> reply := f())
(*  done *)
    ;
  Unixqueue.run esys;
  !reply
;;

let mk_mapping prog vers prot port =
  XV_struct 
    [ "prog", XV_uint prog;
      "vers", XV_uint vers;
      "prot", XV_uint (if prot = Tcp then uint4_of_int 6 else uint4_of_int 17);
      "port", XV_uint (uint4_of_int port)
    ]
;;


let rec dest_pmaplist l =
  try
    begin match l with
      XV_union_over_enum ("TRUE", s) ->
	begin match s with
	  XV_struct [ "map", XV_struct [ "prog", XV_uint prog;
					 "vers", XV_uint vers;
					 "prot", XV_uint prot;
					 "port", XV_uint port 
				       ];
		      "next", l'
		    ] 
	  ->
	    (prog, vers, prot, port) :: dest_pmaplist l'
	end
    | XV_union_over_enum ("FALSE", _) ->
	[]
    end
  with
    Match_failure _ -> failwith "dest_pmaplist"
;;


(*** The procedures of the port mapper ***)

let pm2_null server =
  ignore(exec_pm2 "NULL" XV_void server);
  ()
;;


let pm2_set server prog vers prot port =
  let reply =
    exec_pm2 
      "SET" 
      (mk_mapping prog vers prot port)
      server
  in
  reply = xv_true
;;


let pm2_unset server prog vers prot port =
  let reply =
    exec_pm2 
      "UNSET" 
      (mk_mapping prog vers prot port)
      server
  in
  reply = xv_true
;;


let pm2_getport server prog vers prot port =
  let reply =
    exec_pm2 
      "GETPORT" 
      (mk_mapping prog vers prot port)
      server
  in
  match reply with
    XV_uint n -> int_of_uint4 n
  | _         -> failwith "pm2_getport"
;;


let pm2_dump server =
  let reply =
    exec_pm2 
      "DUMP" 
      XV_void
      server
  in
  dest_pmaplist reply
;;


let pm2_dump_as_int server =
  List.map (fun (a,b,c,d) -> int_of_uint4 a,
                             int_of_uint4 b,
                             int_of_uint4 c,
                             int_of_uint4 d) 
           (pm2_dump server)
;;

                             
